class Game{
	constructor(){
		var that = this;
		
		this.canvas = document.getElementById("canvas");
		this.context = this.canvas.getContext("2d");
		
		this.quality = 1;
		this.updateDelta = 1;
		
		this.ratio = new FixedRatio(1600/3, 900/3, function(left, top, width, height){
			that.canvas.style.left = left + "px";
			that.canvas.style.top = top + "px";
			
			that.canvas.style.width = width + "px";
			that.canvas.style.height = height + "px";
			
			that.canvas.width = width * that.quality;
			that.canvas.height = height * that.quality;
			
			that.context.imageSmoothingEnabled = false;
			
			document.getElementById("fullscreen").hidden = innerWidth >= screen.width - 2 && innerHeight >= screen.height - 2;
		});
		
		this.loop = new GenericLoop(1/60, this.update.bind(this), this.draw.bind(this), function(fps){
			document.getElementById("fps").innerText = fps + " fps";
		});
		
		this.scene = new Scene();
		this.input = new InputHandler(this.ratio);
		this.physics = new BoxPhysics2D();
		this.animations = new Animations();
		
		
		this.atlas = {
			projectiles: {
				bolt: [ 1, 1, 5, 11 ],
				pressureBolt: [ 7, 1, 5, 11 ],
				mageBlade: [ 13, 6, 4, 2 ],
				flamethrower: [
					[ 13, 1, 3, 3 ],
					[ 17, 1, 3, 3 ],
					[ 21, 1, 5, 4 ],
					[ 27, 1, 5, 4 ]
				],
				fireBomb: [ 18, 6, 5, 5 ],
				pressureMine: [ 24, 23, 5, 5 ]
			},
			
			particles: {
				reticle: [ 16, 33, 13, 13 ],
				boltTrail: [ 16, 15, 5, 5 ],
				wizardTrail: [ 24, 15, 5, 5 ],
				
				explosion: {
					orange: [
						[ 35,   3, 32, 32 ],
						[ 35,  38, 32, 32 ],
						[ 35,  73, 32, 32 ],
						[ 35, 108, 32, 32 ],
						[ 35, 143, 32, 32 ],
						[ 35, 178, 32, 32 ],
						[ 35, 213, 32, 32 ],
						[ 35, 248, 32, 32 ],
						[ 35, 283, 32, 32 ],
						[ 35, 318, 32, 32 ]
					],
					
					green: [
						[ 70,   3, 32, 32 ],
						[ 70,  38, 32, 32 ],
						[ 70,  73, 32, 32 ],
						[ 70, 108, 32, 32 ],
						[ 70, 143, 32, 32 ],
						[ 70, 178, 32, 32 ],
						[ 70, 213, 32, 32 ],
						[ 70, 248, 32, 32 ],
						[ 70, 283, 32, 32 ],
						[ 70, 318, 32, 32 ]
					]
				}
			},
			
			players: {
				wizard: [
					[ 29, 1, 6, 6 ],
					[ 29, 8, 6, 6 ],
					[ 29, 15, 6, 6 ]
				],
				
				infantry: {
					pistol: {
						left:  [ [ 8, 1, 6, 6 ], [ 15, 1, 6, 6 ], [ 22, 1, 6, 6 ] ],
						right: [ [ 8, 8, 6, 6 ], [ 15, 8, 6, 6 ], [ 22, 8, 6, 6 ] ]
					},
					
					pistols: {
						left:  [ [ 8, 15, 6, 6 ], [ 15, 15, 6, 6 ], [ 22, 15, 6, 6 ] ],
					},
					
					launcher: {
						left:  [ [ 9, 22, 7, 8 ], [ 17, 22, 7, 8 ], [ 25, 22, 7, 8 ] ],
						right: [ [ 9, 31, 7, 8 ], [ 17, 31, 7, 8 ], [ 25, 31, 7, 8 ] ]
					}
				},
				
				pilot: {
					pistol: {
						left:  [ [ 1, 1, 6, 6 ] ],
						right: [ [ 1, 8, 6, 6 ] ]
					},
					
					pistols: {
						left:  [ [ 1, 15, 6, 6 ] ]
					},
					
					launcher: {
						left:  [ [ 1, 22, 7, 8 ] ],
						right: [ [ 1, 31, 7, 8 ] ]
					}
				}
			},
			
			gui: {
				wave: [ 1, 1, 34, 8 ],
				
				numbers: [
					[ 1, 10, 6, 8 ],
					[ 8, 10, 6, 8 ],
					[ 15, 10, 6, 8 ],
					[ 1, 19, 6, 8 ],
					[ 8, 19, 6, 8 ],
					[ 15, 19, 6, 8 ],
					[ 1, 28, 6, 8 ],
					[ 8, 28, 6, 8 ],
					[ 15, 28, 6, 8 ],
					[ 1, 37, 6, 8 ]
				],
				
				hp: [
					[ 36, 1, 12, 12 ],
					[ 49, 1, 12, 12 ],
					[ 62, 1, 12, 12 ]
				],
				
				spells: [
					[ 62, 27, 12, 12 ],
					[ 36, 27, 12, 12 ],
					[ 49, 27, 12, 12 ],
					[ 36, 40, 12, 12 ],
					[ 49, 40, 12, 12 ],
					[ 36, 53, 12, 12 ],
					[ 49, 53, 12, 12 ]
				],
				
				spellCooldown: [ 62, 40, 12, 12 ],
				spellShadow: [ 62, 53, 12, 12 ],
				
				font: {
					white: [
						[1,1,4,7],
						[6,1,4,7],
						[11,1,4,7],
						[16,1,4,7],
						[21,1,4,7],
						[1,9,4,7],
						[6,9,4,7],
						[11,9,4,7],
						[16,9,5,7],
						[23,9,4,7],
						[1,17,4,7],
						[6,17,4,7],
						[11,17,5,7],
						[17,17,4,7],
						[22,17,4,7],
						[1,25,4,7],
						[6,25,4,7],
						[11,25,4,7],
						[16,25,4,7],
						[21,25,5,7],
						[1,33,4,7],
						[6,33,5,7],
						[12,33,5,7],
						[18,33,5,7],
						[24,33,5,7],
						[1,41,5,7],
						[7,41,4,7],
						[12,41,3,7],
						[16,41,4,7],
						[21,41,4,7],
						[26,41,4,7],
						[1,49,4,7],
						[6,49,4,7],
						[11,49,4,7],
						[16,49,4,7],
						[21,49,4,7],
						[26,49,1,7],
						[28,49,1,7],
						[1,57,4,7],
						[6,57,2,7],
						[9,57,3,7],
						[13,57,1,7],
						[15,57,1,7],
						[17,57,2,7],
						[20,57,3,7],
						[24,57,3,7],
						[28,13,5,43],
						[1,65,3,7],
						[5,65,5,7],
						[11,65,5,7],
						[17,65,5,7],
						[23,65,5,7],
						[26,1,5,7]
					]
				},
				fontMap: "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.!?,\"':;+-=*_\\|/ "
			}
		};
		
		
		this.textures = {
			players: Texture("assets/players.png"),
			bounds: Texture("assets/bounds.png"),
			
			alien1: Texture(Structures.alien1.base64),
			alien2: Texture(Structures.alien2.base64),
			alien3: Texture(Structures.alien3.base64),
			alien4: Texture(Structures.alien4.base64),
			alien5: Texture(Structures.alien5.base64),
			alien6: Texture(Structures.alien6.base64),
			alien7: Texture(Structures.alien7.base64),
			
			forcefield: Texture("assets/forcefield.png"),
			forcefield2: Texture("assets/forcefield2.png"),
			
			world: {
				stars1: Texture("assets/stars-1.png"),
				stars2: Texture("assets/stars-2.png")
			},
			
			projectiles: Texture("assets/projectiles.png"),
			fonts: Texture("assets/fonts.png"),
			title: Texture("assets/title.png"),
			gui: Texture("assets/gui.png"),
			
			wizardHead: Texture("assets/wizardhead.png"),
			soldierHead: Texture("assets/soldierhead.png")
		};
		
		
		// Monitor loading of all images
		var b64Count = 0;
		
		for(var i in this.textures){
			if(this.textures[i] instanceof Image /*&& this.textures[i].src.startsWith("data:")*/){
				b64Count++;
				
				this.textures[i].onload = function(){
					if(--b64Count === 0){
						that.initScript.call(that, that);
						that.loop.start();
					}
				};
			}
		}
		
		
		// Reset the scene
		this.scale = 2;
		this.shakeTime = 0;
		this.shakeTimeMax = 0;
		this.shakeIntensity = 0;
		this.initScript = IntroScript;
		
		this.sceneX = 330/2;
		this.sceneY = 330/2;
		this.desirX = this.sceneX;
		this.desirY = this.sceneY;
		
		this.scene.setRule("particles", 1);
	}
	
	
	stateUpdate(){}
	stateDraw(){}
	
	
	setQuality(quality){
		this.quality = quality;
		this.ratio.onChange(this.ratio.screenLeft, this.ratio.screenRight, this.ratio.screenWidth, this.ratio.screenHeight);
	}
	
	
	canShowParticle(factor){
		return this.loop.frame % factor === 0;
	}
	
	
	playTrack(track){
		if(this.updateDelta !== 1) return;
		
		var audio;
		
		switch(track){
			case "explode":
				audio = Tracks.explode[Math.floor(Math.random() * Tracks.explode.length)];
				break;
			
			case "shoot":
				audio = Tracks.shoot[Math.floor(Math.random() * Tracks.shoot.length)];
				break;
			
			case "cast":
				audio = Tracks.cast[Math.floor(Math.random() * Tracks.cast.length)];
				break;
			
			default:
				return;
		}
		
		audio.play();
	}
	
	
	shake(intensity, time){
		if(intensity>this.shakeIntensity){
			this.shakeTime = this.shakeTimeMax = time;
			this.shakeIntensity = intensity;
		}
	}


	update(delta){
		var atlas = this.atlas;
		
		this.scene.updateObjects(delta * this.updateDelta, this.input);
		this.scene.updateParticles(delta * this.updateDelta);
		this.animations.update(delta);
		
		var ease = 16;
		this.sceneX += (this.desirX - this.sceneX) / ease;
		this.sceneY += (this.desirY - this.sceneY) / ease;
		
		if(this.shakeTime>0) this.shakeTime -= delta;
		else this.shakeTime = 0, this.shakeIntensity = 0;
		
		this.stateUpdate(delta);
	}
	
	
	draw(delta){
		var that = this,
			frame = this.loop.frame;
		
		// Prepare the canvas
		this.context.save();
		this.context.scale(this.ratio.scale * this.scale * this.quality, this.ratio.scale * this.scale * this.quality);
		this.context.clearRect(0, 0, this.ratio.width / this.scale, this.ratio.height / this.scale);
		
		// Shake the screen
		if(this.shakeTime > 0){
			var intensity = this.shakeTime / this.shakeTimeMax * this.shakeIntensity,
				width  = this.ratio.width  / this.scale,
				height = this.ratio.height / this.scale,
				
				timeR = 0.1,
				timeX = 0.14,
				timeY = 0.14,
				
				bounceR = rebound(this.shakeTime, timeR) / timeR * 2 - 1,
				bounceX = rebound(this.shakeTime, timeX) / timeX * 2 - 1,
				bounceY = rebound(this.shakeTime + timeY/2, timeY) / timeY * 2 - 1,
				
				shakeR = intensity * bounceR * 10,
				shakeX = intensity * bounceX * 10,
				shakeY = intensity * bounceY * 10;
			
			this.context.translate(width/2 + shakeX, height/2 + shakeY);
			this.context.rotate(shakeR/180*Math.PI);
			this.context.translate(-this.sceneX, -this.sceneY);
		}
		
		else{
			this.context.translate(-this.sceneX + this.ratio.width / 2 / this.scale, -this.sceneY + this.ratio.height / 2 / this.scale);
		}
		
		
		// Draw the background
		var world = this.scene.getRule("world");
		if(world==="tundra") tilemap(this.textures.world.tundra, 0, 0, 32, 32, 1);
		if(world==="space"){
			tilemap(this.textures.world.stars1, this.sceneX * 0.8, (this.sceneY + this.loop.frame/60) * 0.8, 800, 600, 0.2);
			tilemap(this.textures.world.stars2, this.sceneX * 0.78, (this.sceneY + this.loop.frame/60) * 0.78, 800, 600, 0.5);
		}
		
		this.scene.drawParticlesBg(this.context, this.textures, this.atlas);
		
		// Draw the midground
		this.context.globalAlpha = 1;
		this.scene.drawObjects(this.context, this.textures, this.atlas);
		this.scene.drawParticlesFg(this.context, this.textures, this.atlas);
		
		// Draw the foreground
		if(this.scene.getRule("isSnowing")){
			tilemap(this.textures.world.snow, frame, -frame/20, 200, 150, 1);
			tilemap(this.textures.world.snow, frame/2 - 100, frame/20, 200, 150, 1);
		}
		
		// Draw the bounds
		if(this.scene.getRule("useBounds")){
			var scale = 1,
			width = scale * 330,
			height = scale * 330,
			index = Math.floor(this.loop.frame/4) % 3;
			
			var screenX = this.sceneX - this.ratio.width/2/this.scale,
				screenY = this.sceneY - this.ratio.height/2/this.scale,
				padding = true || this.shakeTime > 0 ? 100 : 0;
			
			this.context.globalAlpha = 0.65;
			this.context.drawImage(this.textures.bounds, 0, 0, 1, 1, screenX, screenY - padding, this.ratio.width/this.scale, -screenY + padding);
			this.context.drawImage(this.textures.bounds, 0, 0, 1, 1, screenX, height,  this.ratio.width/this.scale, this.sceneY - height + this.ratio.height/2/this.scale + padding);
			this.context.drawImage(this.textures.bounds, 0, 0, 1, 1, screenX - padding, screenY, -screenX + padding, this.ratio.height/this.scale);
			this.context.drawImage(this.textures.bounds, 0, 0, 1, 1, width,   screenY, this.sceneX - width + this.ratio.width/2/this.scale + padding, this.ratio.height/this.scale);
			
			this.context.drawImage(this.textures.bounds, 0, index * height, width, height, 0, 0, width, height);
			// this.context.drawImage(this.textures.bounds, 0, index * height, width, height, -1/this.scale, -1/this.scale, width + 2/this.scale, height + 2/this.scale);
			this.context.globalAlpha = 1;
			
		}
		
		this.context.restore();
		
		this.context.save();
		this.context.scale(this.ratio.scale * this.scale * this.quality, this.ratio.scale * this.scale * this.quality);
		this.context.globalAlpha = 1;
		
		this.stateDraw(this.context, this.textures, this.atlas);
		this.animations.draw();
		this.context.restore();
		
		function tilemap(texture, offsetX, offsetY, textureWidth, textureHeight, scale){
			var width  = that.ratio.width  / (textureWidth  * scale)  / that.scale,
				height = that.ratio.height / (textureHeight  * scale) / that.scale,
				sx = Math.floor((that.sceneX - that.ratio.width /2/that.scale) / textureWidth ) - 2,
				sy = Math.floor((that.sceneY - that.ratio.height/2/that.scale) / textureHeight) - 2,
				ex = sx + width + 4,
				ey = sy + height + 4;
			
			for(var y=sy; y<ey; y++){
				for(var x=sx; x<ex; x++){
					that.context.drawImage(
						texture,
						x * (textureWidth  * scale)  + (offsetX % (textureWidth  * scale)),
						y * (textureHeight * scale) + (offsetY % (textureHeight * scale)),
						textureWidth  * scale + 1/that.scale,
						textureHeight * scale + 1/that.scale
					);
				}
			}
		}
	}
}
